"""
 @file
 @brief This file loads the Addtotimeline dialog (i.e add several clips in the timeline)
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <olivier@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import uuid
from operator import itemgetter
from random import shuffle, randint, uniform

from PyQt5.QtCore import QLocale
from PyQt5.QtWidgets import QDialog
from PyQt5.QtGui import QIcon

from classes import info, ui_util, time_parts
from classes.logger import log
from classes.query import Clip, Transition
from classes.app import get_app
from classes.metrics import track_metric_screen
from classes.clip_utils import apply_file_caption_to_clip
from windows.views.add_to_timeline_treeview import TimelineTreeView

import openshot
import json


class AddToTimeline(QDialog):
    """ Add To timeline Dialog """

    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'add-to-timeline.ui')

    def _select_added_items(self, win, added_clip_ids):
        """Select only the newly added clips, matching timeline drag/drop behavior."""
        if not win or not added_clip_ids:
            return

        timeline_view = getattr(win, "timeline", None)
        for idx, clip_id in enumerate(added_clip_ids):
            if not clip_id:
                continue
            if timeline_view and hasattr(timeline_view, "AddSelectionJS"):
                timeline_view.AddSelectionJS(str(clip_id), "clip", idx == 0)
            else:
                win.addSelection(str(clip_id), "clip", clear_existing=(idx == 0))

        files_model = getattr(win, "files_model", None)
        if files_model:
            selection_model = getattr(files_model, "selection_model", None)
            if selection_model:
                selection_model.clearSelection()
            list_selection_model = getattr(files_model, "list_selection_model", None)
            if list_selection_model:
                list_selection_model.clearSelection()

        if timeline_view and hasattr(timeline_view, "setFocus"):
            timeline_view.setFocus()
        if timeline_view and hasattr(timeline_view, "geometry"):
            timeline_geometry = getattr(timeline_view, "geometry", None)
            if hasattr(timeline_geometry, "mark_dirty"):
                timeline_geometry.mark_dirty()
        if timeline_view and hasattr(timeline_view, "update"):
            timeline_view.update()

    def btnMoveUpClicked(self, checked):
        """Callback for move up button click"""
        log.info("btnMoveUpClicked")

        # Get selected file
        files = self.treeFiles.timeline_model.files

        selected_index = None
        if self.treeFiles.selected:
            selected_index = self.treeFiles.selected.row()

        # Ignore if empty files or no selection
        if not files or selected_index is None:
            return

        # Check if selected_index is within valid range
        if 0 <= selected_index < len(files):
            # New index
            new_index = max(selected_index - 1, 0)

            # Remove item and move it
            files.insert(new_index, files.pop(selected_index))
        else:
            log.warning(f"Invalid selected_index: {selected_index}, list length: {len(files)}")
            return

        # Refresh tree
        self.treeFiles.refresh_view()

        # Select new position
        idx = self.treeFiles.timeline_model.model.index(new_index, 0)
        self.treeFiles.setCurrentIndex(idx)

    def btnMoveDownClicked(self, checked):
        """Callback for move up button click"""
        log.info("btnMoveDownClicked")

        # Get selected file
        files = self.treeFiles.timeline_model.files

        selected_index = None
        if self.treeFiles.selected:
            selected_index = self.treeFiles.selected.row()

        # Ignore if empty files or no selection
        if not files or selected_index is None:
            return

        # Check if selected_index is within valid range
        if 0 <= selected_index < len(files):
            # New index
            new_index = min(selected_index + 1, len(files) - 1)

            # Remove item and move it
            files.insert(new_index, files.pop(selected_index))
        else:
            log.warning(f"Invalid selected_index: {selected_index}, list length: {len(files)}")
            return

        # Refresh tree
        self.treeFiles.refresh_view()

        # Select new position
        idx = self.treeFiles.timeline_model.model.index(new_index, 0)
        self.treeFiles.setCurrentIndex(idx)

    def btnShuffleClicked(self, checked):
        """Callback for move up button click"""
        log.info("btnShuffleClicked")

        # Shuffle files
        shuffle(self.treeFiles.timeline_model.files)

        # Refresh tree
        self.treeFiles.refresh_view()

    def btnRemoveClicked(self, checked):
        """Callback for move up button click"""
        log.info("btnRemoveClicked")

        # Get selected file
        files = self.treeFiles.timeline_model.files

        selected_index = None
        if self.treeFiles.selected:
            selected_index = self.treeFiles.selected.row()

        # Ignore if empty files
        if not files or selected_index is None:
            return

        # Remove item
        files.pop(selected_index)

        # Refresh tree
        self.treeFiles.refresh_view()

        # Select next item (if any)
        new_index = max(len(files) - 1, 0)

        # Select new position
        idx = self.treeFiles.timeline_model.model.index(new_index, 0)
        self.treeFiles.setCurrentIndex(idx)

        # Update total
        self.updateTotal()

    def accept(self):
        """ Ok button clicked """
        log.info('accept')

        # Transaction id to group all updates together
        tid = str(uuid.uuid4())
        get_app().updates.transaction_id = tid

        # Get settings from form
        start_position = self.txtStartTime.value()
        track_num = self.cmbTrack.currentData()
        fade_value = self.cmbFade.currentData()
        fade_length = self.txtFadeLength.value()
        transition_path = self.cmbTransition.currentData()
        transition_length = self.txtTransitionLength.value()
        image_length = self.txtImageLength.value()
        zoom_value = self.cmbZoom.currentData()

        # Init position
        position = start_position

        random_transition = False
        if transition_path == "random":
            random_transition = True

        # Get frames per second
        fps = get_app().project.get("fps")
        fps_float = float(fps["num"]) / float(fps["den"])

        # Track added clip IDs for auto-selection
        added_clip_ids = []

        # Loop through each file (in the current order)
        for file in self.treeFiles.timeline_model.files:
            # Create a clip
            clip = Clip()
            clip.data = {}

            # Get file name
            filename = os.path.basename(file.data["path"])

            # Convert path to the correct relative path (based on this folder)
            file_path = file.absolute_path()

            # Create clip object for this file
            c = openshot.Clip(file_path)

            # Append missing attributes to Clip JSON
            new_clip = json.loads(c.Json())
            new_clip["position"] = position
            new_clip["layer"] = track_num
            new_clip["file_id"] = file.id
            new_clip["title"] = file.data.get("name", filename)
            new_clip["reader"] = file.data

            # Skip any clips that are missing a 'reader' attribute
            # TODO: Determine why this even happens, as it shouldn't be possible
            if not new_clip.get("reader"):
                continue  # Skip to next file

            # If the source file has stored caption text, attach a Caption effect to this new clip.
            apply_file_caption_to_clip(new_clip, file)

            # Check for optional start and end attributes
            start_time = 0
            end_time = new_clip["reader"]["duration"]

            if 'start' in file.data:
                start_time = file.data['start']
                new_clip["start"] = start_time
            if 'end' in file.data:
                end_time = file.data['end']
                new_clip["end"] = end_time

            # Adjust clip duration, start, and end
            new_clip["duration"] = new_clip["reader"]["duration"]
            if file.data["media_type"] == "image":
                end_time = image_length
                new_clip["end"] = end_time
            else:
                new_clip["end"] = end_time

            # Adjust Fade of Clips (if no transition is chosen)
            if not transition_path:
                if fade_value is not None:
                    # Overlap this clip with the previous one (if any)
                    position = max(start_position, new_clip["position"] - fade_length)
                    new_clip["position"] = position

                if fade_value in ['Fade In', 'Fade In & Out']:
                    start = openshot.Point(round(start_time * fps_float) + 1, 0.0, openshot.BEZIER)
                    start_object = json.loads(start.Json())
                    end = openshot.Point(
                        min(
                            round((start_time + fade_length) * fps_float) + 1,
                            round(end_time * fps_float) + 1
                            ),
                        1.0,
                        openshot.BEZIER)
                    end_object = json.loads(end.Json())
                    new_clip['alpha']["Points"].append(start_object)
                    new_clip['alpha']["Points"].append(end_object)

                if fade_value in ['Fade Out', 'Fade In & Out']:
                    start = openshot.Point(
                        max(
                            round((end_time * fps_float) + 1) - (round(fade_length * fps_float) + 1),
                            round(start_time * fps_float) + 1
                            ),
                        1.0,
                        openshot.BEZIER)
                    start_object = json.loads(start.Json())
                    end = openshot.Point(
                        round(end_time * fps_float) + 1,
                        0.0,
                        openshot.BEZIER)
                    end_object = json.loads(end.Json())
                    new_clip['alpha']["Points"].append(start_object)
                    new_clip['alpha']["Points"].append(end_object)

            # Adjust zoom amount
            if zoom_value is not None:
                # Location animation
                if zoom_value == "Random":
                    animate_start_x = uniform(-0.5, 0.5)
                    animate_end_x = uniform(-0.15, 0.15)
                    animate_start_y = uniform(-0.5, 0.5)
                    animate_end_y = uniform(-0.15, 0.15)

                    # Scale animation
                    start_scale = uniform(0.5, 1.5)
                    end_scale = uniform(0.85, 1.15)

                elif zoom_value == "Zoom In":
                    animate_start_x = 0.0
                    animate_end_x = 0.0
                    animate_start_y = 0.0
                    animate_end_y = 0.0

                    # Scale animation
                    start_scale = 1.0
                    end_scale = 1.25

                elif zoom_value == "Zoom Out":
                    animate_start_x = 0.0
                    animate_end_x = 0.0
                    animate_start_y = 0.0
                    animate_end_y = 0.0

                    # Scale animation
                    start_scale = 1.25
                    end_scale = 1.0

                # Add keyframes
                start = openshot.Point(round(start_time * fps_float) + 1, start_scale, openshot.BEZIER)
                start_object = json.loads(start.Json())
                end = openshot.Point(round(end_time * fps_float) + 1, end_scale, openshot.BEZIER)
                end_object = json.loads(end.Json())
                new_clip["gravity"] = openshot.GRAVITY_CENTER
                new_clip["scale_x"]["Points"].append(start_object)
                new_clip["scale_x"]["Points"].append(end_object)
                new_clip["scale_y"]["Points"].append(start_object)
                new_clip["scale_y"]["Points"].append(end_object)

                # Add keyframes
                start_x = openshot.Point(round(start_time * fps_float) + 1, animate_start_x, openshot.BEZIER)
                start_x_object = json.loads(start_x.Json())
                end_x = openshot.Point(round(end_time * fps_float) + 1, animate_end_x, openshot.BEZIER)
                end_x_object = json.loads(end_x.Json())
                start_y = openshot.Point(round(start_time * fps_float) + 1, animate_start_y, openshot.BEZIER)
                start_y_object = json.loads(start_y.Json())
                end_y = openshot.Point(round(end_time * fps_float) + 1, animate_end_y, openshot.BEZIER)
                end_y_object = json.loads(end_y.Json())
                new_clip["gravity"] = openshot.GRAVITY_CENTER
                new_clip["location_x"]["Points"].append(start_x_object)
                new_clip["location_x"]["Points"].append(end_x_object)
                new_clip["location_y"]["Points"].append(start_y_object)
                new_clip["location_y"]["Points"].append(end_y_object)

            if transition_path:
                # Add transition for this clip (if any)
                # Open up QtImageReader for transition Image
                if random_transition:
                    random_index = randint(0, len(self.transitions) - 1)
                    transition_path = self.transitions[random_index]

                # Get reader for transition
                transition_reader = openshot.QtImageReader(transition_path)

                brightness = openshot.Keyframe()
                brightness.AddPoint(1, 1.0, openshot.BEZIER)
                brightness.AddPoint(
                    round(
                        min(transition_length, end_time - start_time)
                        * fps_float
                        ) + 1,
                    -1.0,
                    openshot.BEZIER)
                contrast = openshot.Keyframe(3.0)

                # Create transition dictionary
                transitions_data = {
                    "layer": track_num,
                    "title": "Transition",
                    "type": "Mask",
                    "start": 0,
                    "end": min(transition_length, end_time - start_time),
                    "brightness": json.loads(brightness.Json()),
                    "contrast": json.loads(contrast.Json()),
                    "reader": json.loads(transition_reader.Json()),
                    "replace_image": False
                }

                # Overlap this clip with the previous one (if any)
                position = max(start_position, position - transition_length)
                transitions_data["position"] = position
                new_clip["position"] = position

                # Create transition
                tran = Transition()
                tran.data = transitions_data
                tran.save()

            # Save Clip
            clip.data = new_clip
            clip.save()
            added_clip_ids.append(clip.data.get("id"))

            # Increment position by length of clip
            position += (end_time - start_time)

        # Clear transaction
        get_app().updates.transaction_id = None

        win = get_app().window

        # Ensure project duration grows to include all newly-added items.
        timeline_view = getattr(win, "timeline", None)
        extend_timeline = getattr(timeline_view, "_extend_timeline_to_fit_items", None)
        if callable(extend_timeline):
            try:
                extend_timeline()
            except Exception:
                log.warning("Failed to extend timeline after Add to Timeline", exc_info=1)

        # Auto-select newly added clips, like timeline drag/drop does.
        self._select_added_items(win, added_clip_ids)

        # Accept dialog
        super(AddToTimeline, self).accept()

    def ImageLengthChanged(self, value):
        """Handle callback for image length being changed"""
        self.updateTotal()

    def updateTotal(self):
        """Calculate the total length of what's about to be added to the timeline"""
        fade_value = self.cmbFade.currentData()
        fade_length = self.txtFadeLength.value()
        transition_path = self.cmbTransition.currentData()
        transition_length = self.txtTransitionLength.value()

        total = 0.0
        for file in self.treeFiles.timeline_model.files:
            # Adjust clip duration, start, and end
            duration = file.data["duration"]
            if file.data["media_type"] == "image":
                duration = self.txtImageLength.value()

            if total != 0.0:
                # Don't subtract time from initial clip
                if not transition_path:
                    # No transitions
                    if fade_value is not None:
                        # Fade clip - subtract the fade length
                        duration -= fade_length
                else:
                    # Transition
                    duration -= transition_length

            # Append duration to total
            total += duration

        # Get frames per second
        fps = get_app().project.get("fps")

        # Update label
        total_parts = time_parts.secondsToTime(total, fps["num"], fps["den"])
        timestamp = "%s:%s:%s:%s" % (total_parts["hour"], total_parts["min"], total_parts["sec"], total_parts["frame"])
        self.lblTotalLengthValue.setText(timestamp)

    def reject(self):
        """ Cancel button clicked """
        log.info('reject')

        # Accept dialog
        super(AddToTimeline, self).reject()

    def __init__(self, files=None, position=0.0):
        # Create dialog class
        QDialog.__init__(self)

        # Load UI from Designer
        ui_util.load_ui(self, self.ui_path)

        # Init UI
        ui_util.init_ui(self)

        # Get translation object
        self.app = get_app()
        _ = self.app._tr

        # Get settings
        self.settings = self.app.get_settings()

        # Track metrics
        track_metric_screen("add-to-timeline-screen")

        # Add custom treeview to window
        self.treeFiles = TimelineTreeView(self)
        self.vboxTreeParent.insertWidget(0, self.treeFiles)

        # Update data in model
        self.treeFiles.timeline_model.update_model(files)

        # Init start position
        self.txtStartTime.setValue(position)

        # Init default image length
        self.txtImageLength.setValue(self.settings.get("default-image-length"))
        self.txtImageLength.valueChanged.connect(self.updateTotal)
        self.cmbTransition.currentIndexChanged.connect(self.updateTotal)
        self.cmbFade.currentIndexChanged.connect(self.updateTotal)
        self.txtFadeLength.valueChanged.connect(self.updateTotal)
        self.txtTransitionLength.valueChanged.connect(self.updateTotal)

        # Find display track number
        all_tracks = get_app().project.get("layers")
        display_count = len(all_tracks)
        for track in reversed(sorted(all_tracks, key=itemgetter('number'))):
            # Add to dropdown
            track_name = track.get('label') or _("Track %s") % QLocale().toString(display_count)
            self.cmbTrack.addItem(track_name, track.get('number'))
            display_count -= 1

        # Add all fade options
        self.cmbFade.addItem(_('None'), None)
        self.cmbFade.addItem(_('Fade In'), 'Fade In')
        self.cmbFade.addItem(_('Fade Out'), 'Fade Out')
        self.cmbFade.addItem(_('Fade In & Out'), 'Fade In & Out')

        # Add all zoom options
        self.cmbZoom.addItem(_('None'), None)
        self.cmbZoom.addItem(_('Random'), 'Random')
        self.cmbZoom.addItem(_('Zoom In'), 'Zoom In')
        self.cmbZoom.addItem(_('Zoom Out'), 'Zoom Out')

        # Add all transitions
        transitions_dir = os.path.join(info.PATH, "transitions")
        common_dir = os.path.join(transitions_dir, "common")
        extra_dir = os.path.join(transitions_dir, "extra")
        transition_groups = [{"type": "common", "dir": common_dir, "files": os.listdir(common_dir)},
                             {"type": "extra", "dir": extra_dir, "files": os.listdir(extra_dir)}]

        self.cmbTransition.addItem(_('None'), None)
        self.cmbTransition.addItem(_('Random'), 'random')
        self.transitions = []
        for group in transition_groups:
            dir = group["dir"]
            files = group["files"]

            for filename in sorted(files):
                path = os.path.join(dir, filename)
                fileBaseName = os.path.splitext(filename)[0]

                # Skip hidden files (such as .DS_Store, etc...)
                if filename[0] == "." or "thumbs.db" in filename.lower():
                    continue

                # split the name into parts (looking for a number)
                suffix_number = None
                name_parts = fileBaseName.split("_")
                if name_parts[-1].isdigit():
                    suffix_number = name_parts[-1]

                # get name of transition
                trans_name = fileBaseName.replace("_", " ").capitalize()

                # replace suffix number with placeholder (if any)
                if suffix_number:
                    trans_name = trans_name.replace(suffix_number, "%s")
                    trans_name = _(trans_name) % QLocale().toString(int(suffix_number))
                else:
                    trans_name = _(trans_name)

                # Check for thumbnail path (in build-in cache)
                thumb_path = os.path.join(info.IMAGES_PATH, "cache",  "{}.png".format(fileBaseName))

                # Check built-in cache (if not found)
                if not os.path.exists(thumb_path):
                    # Check user folder cache
                    thumb_path = os.path.join(info.CACHE_PATH, "{}.png".format(fileBaseName))

                # Add item
                self.transitions.append(path)
                self.cmbTransition.addItem(QIcon(thumb_path), _(trans_name), path)

        # Connections
        self.btnMoveUp.clicked.connect(self.btnMoveUpClicked)
        self.btnMoveDown.clicked.connect(self.btnMoveDownClicked)
        self.btnShuffle.clicked.connect(self.btnShuffleClicked)
        self.btnRemove.clicked.connect(self.btnRemoveClicked)
        self.btnBox.accepted.connect(self.accept)
        self.btnBox.rejected.connect(self.reject)

        # Update total
        self.updateTotal()



"""
Timeline backend package helpers.

Ensures legacy ``classes`` imports work when OpenShot is installed under
the ``openshot_qt`` package.
"""

import os
import sys

try:
    import classes  # noqa: F401
except ImportError:
    # When running from a source checkout (for example via unittest discovery),
    # ensure the repository's ``src`` directory is on sys.path so ``classes``
    # can be imported without installing the package.
    checkout_src = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, os.pardir)
    )
    if (
        os.path.isdir(os.path.join(checkout_src, "classes"))
        and checkout_src not in sys.path
    ):
        sys.path.insert(0, checkout_src)

    try:
        import classes  # noqa: F401
    except ImportError:
        try:
            import openshot_qt

            # Prefer OPENSHOT_PATH if upstream defines it, else use package dir
            pkg_dir = getattr(openshot_qt, "OPENSHOT_PATH", None) or os.path.dirname(
                openshot_qt.__file__
            )
            if pkg_dir and pkg_dir not in sys.path:
                sys.path.insert(0, pkg_dir)

            import classes  # noqa: F401
        except Exception:
            # Let the original ImportError surface from downstream imports
            pass



"""
 @file
 @brief This file connects to libopenshot and logs debug messages (if debug preference enabled)
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from threading import Thread
from classes import info
from classes.logger import log
from classes.app import get_app
import openshot
import os
import zmq


class LoggerLibOpenShot(Thread):
    def __init__(self):
        super().__init__()
        self.daemon = True
        self.running = False
        self.context = None
        self.socket = None


    def kill(self):
        self.running = False
        log.info('Shutting down libopenshot logger')

    def run(self):
        # Running
        self.running = True

        # Get settings
        s = get_app().get_settings()

        # Get port from settings
        port = s.get("debug-port")
        debug_enabled = s.get("debug-mode")

        # Set port on ZmqLogger singleton
        openshot.ZmqLogger.Instance().Connection("tcp://*:%s" % port)

        # Set filepath for ZmqLogger also
        openshot.ZmqLogger.Instance().Path(os.path.join(info.USER_PATH, 'libopenshot.log'))

        # Enable / Disable logger
        openshot.ZmqLogger.Instance().Enable(debug_enabled)

        # Socket to talk to server
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.SUB)
        self.socket.setsockopt_string(zmq.SUBSCRIBE, '')

        poller = zmq.Poller()
        poller.register(self.socket, zmq.POLLIN)

        log.info("Connecting to libopenshot with debug port: %s" % port)
        self.socket.connect("tcp://localhost:%s" % port)

        while self.running:
            msg = None

            # Receive all debug message sent from libopenshot (if any)
            try:
                socks = dict(poller.poll(1000))
                if socks and socks.get(self.socket) == zmq.POLLIN:
                    msg = self.socket.recv(zmq.NOBLOCK)
                if msg:
                    log.info(msg.strip().decode('UTF-8'))
            except Exception as ex:
                log.warning(ex)

        # Close zmq connection
        if self.context:
            self.context.destroy()
        if self.socket:
            self.socket.close()
        if openshot.ZmqLogger.Instance():
            # Close libopenshot logger
            openshot.ZmqLogger.Instance().Close()



"""
 @file
 @brief This file contains optimized-preview proxy management logic.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import copy
import json
import os
import re
import threading
import uuid
from concurrent.futures import ThreadPoolExecutor

from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QFileDialog

import openshot

from classes import info
from classes.app import get_app
from classes.assets import get_assets_path
from classes.logger import log
from classes.path_utils import absolute_media_path, comparable_media_path
from classes.query import Clip, File, Transition
from classes.thumbnail import (
    GenerateThumbnailFromFrame,
    RoundFrameToThumbnailGrid,
    ThumbnailPathForFrame,
)
from classes.updates import UpdateAction


def dialog_preview_reader_data(file_obj, prefer_proxy=True):
     """Return reader payload for dialog previews, optionally preferring a valid proxy reader."""
     data = getattr(file_obj, "data", {}) if file_obj else {}
     if not isinstance(data, dict):
         data = {}

     file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
     proxy_reader = data.get("proxy_reader")
     if prefer_proxy and isinstance(proxy_reader, dict):
         if proxy_reader.get("missing"):
             proxy_reader = None
         else:
             proxy_path = absolute_media_path(proxy_reader.get("path"))
             if proxy_path and os.path.exists(proxy_path):
                 resolved = copy.deepcopy(proxy_reader)
                 resolved["path"] = proxy_path
                 if file_id:
                     resolved["id"] = file_id
                 return resolved

     resolved = copy.deepcopy(data)
     source_path = absolute_media_path(resolved.get("path"))
     if source_path:
         resolved["path"] = source_path
     if file_id:
         resolved["id"] = file_id
     return resolved


class ProxyService(QObject):
     """Encapsulates optimized-preview proxy generation, linking, and runtime rewrites."""

     proxy_generated = pyqtSignal(str, object, str)
     file_job_changed = pyqtSignal(str)
     queue_changed = pyqtSignal()
     job_updated = pyqtSignal(str, str, int)
     job_finished = pyqtSignal(str, str)
     ACTIVE_STATES = ("queued", "running", "canceling")
     OPTIMIZE_CACHE_MAX_BYTES = 16 * 1024 * 1024
     OPTIMIZE_CACHE_CLEAR_INTERVAL = 120
     MATCHABLE_VIDEO_EXTENSIONS = {
         ".3g2", ".3gp", ".asf", ".avi", ".f4v", ".flv", ".m2t", ".m2ts", ".m4v", ".mkv",
         ".mov", ".mp2", ".mp4", ".mpeg", ".mpg", ".mts", ".mxf", ".ogv", ".rm", ".rmvb",
         ".ts", ".vob", ".webm", ".wmv",
     }
     _PROXY_PREFIX_RE = re.compile(
         r"^(?:proxy|proxies|prox|optimized?|optimised|preview|lowres|low_res|low-res|offline|transcoded?)[\s._-]+",
         re.IGNORECASE,
     )
     _PROXY_SUFFIX_RE = re.compile(
         r"[\s._-]+(?:proxy|proxies|prox|optimized?|optimised|preview|lowres|low_res|low-res|offline|transcoded?)$",
         re.IGNORECASE,
     )
     _RESOLUTION_SUFFIX_RE = re.compile(
         r"[\s._-]+(?:360p|540p|720p|1080p|1440p|2160p|4320p|4k|8k)$",
         re.IGNORECASE,
     )
     _CAMERA_PROXY_SUFFIX_RE = re.compile(r"[\s._-]+p$", re.IGNORECASE)

     def __init__(self, win):
         super().__init__(win)
         self.win = win
         self._executor = None
         self._jobs = {}
         self._lock = threading.RLock()
         self._ensure_executor()
         self.proxy_generated.connect(self._on_proxy_generated)

     def shutdown(self):
         if getattr(self, "_executor", None):
             self._shutdown_executor(self._executor)

     def create_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         self._ensure_executor()
         os.makedirs(self._proxy_root(), exist_ok=True)
         submitted = 0
         linked_existing = 0
         skipped = 0
         for file_obj in files:
             file_id = str(file_obj.id or "")
             if not file_id:
                 continue
             if self.get_active_job_for_file(file_id):
                 log.info("Optimize Preview create_for_files file_id=%s skipped: job already active", file_id)
                 skipped += 1
                 continue
             if self.has_proxy_reader(file_obj) and not self.has_missing_proxy(file_obj):
                 log.info("Optimize Preview create_for_files file_id=%s skipped: already optimized", file_id)
                 skipped += 1
                 continue
             snapshot = copy.deepcopy(file_obj.data or {})
             existing_output_path = self._existing_proxy_output_path(file_id, snapshot)
             if existing_output_path:
                 try:
                     proxy_reader = self._reader_json_for_path(existing_output_path, file_id)
                     self._save_proxy_reader(file_id, proxy_reader)
                     log.info("Optimize Preview create_for_files file_id=%s linked existing output=%s", file_id, existing_output_path)
                     linked_existing += 1
                     continue
                 except Exception as exc:
                     log.warning(
                         "Optimize Preview create_for_files file_id=%s ignored invalid existing output=%s (%s)",
                         file_id,
                         existing_output_path,
                         exc,
                     )
                     try:
                         os.remove(existing_output_path)
                     except Exception as cleanup_exc:
                         log.warning(
                             "Optimize Preview create_for_files file_id=%s failed to remove invalid existing output=%s (%s)",
                             file_id,
                             existing_output_path,
                             cleanup_exc,
                         )
             output_path = self._reserve_proxy_output_path(file_id, snapshot)
             with self._lock:
                 self._jobs[file_id] = {
                     "id": file_id,
                     "status": "queued",
                     "progress": 0,
                     "cancel_requested": False,
                     "output_path": output_path,
                 }
             future = self._executor.submit(self._build_proxy_reader, file_id, snapshot)
             with self._lock:
                 if file_id in self._jobs:
                     self._jobs[file_id]["future"] = future
             future.add_done_callback(lambda fut, fid=file_id: self._emit_proxy_result(fid, fut))
             self._emit_job_change(file_id)
             submitted += 1

         if submitted:
             status = "Optimize Preview: creating {} item(s)".format(submitted)
             if linked_existing:
                 status += ", linked {}".format(linked_existing)
             if skipped:
                 status += ", skipped {}".format(skipped)
             self._show_status(status)
         elif linked_existing:
             status = "Optimize Preview: linked {} item(s)".format(linked_existing)
             if skipped:
                 status += ", skipped {}".format(skipped)
             self._show_status(status, 3000)
         elif skipped:
             self._show_status("Optimize Preview: skipped {} item(s)".format(skipped), 3000)

     def use_existing_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         start_dir = self._proxy_root()
         if not start_dir:
             start_dir = info.HOME_PATH
         translator = getattr(self.win, "_tr", None) or getattr(get_app(), "_tr", lambda text: text)
         selected_folder = QFileDialog.getExistingDirectory(
             self.win,
             translator("Choose optimized preview folder"),
             start_dir,
         )
         if not selected_folder:
             return

         matched = 0
         missing = 0
         invalid = 0
         changed_file_ids = []
         folder_index = self._index_existing_optimized_files(selected_folder)
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 match_path = self._match_existing_optimized_path(file_obj, selected_folder, folder_index)
                 if match_path and os.path.exists(match_path):
                     try:
                         proxy_reader = self._reader_json_for_path(match_path, file_obj.id)
                         matched += 1
                         self._save_proxy_reader(file_obj.id, proxy_reader, apply_runtime=False, emit_job_change=False)
                         changed_file_ids.append(str(file_obj.id or ""))
                     except Exception as exc:
                         log.warning(
                             "Optimize Preview use_existing_for_files file_id=%s skipped invalid existing file=%s (%s)",
                             file_obj.id,
                             match_path,
                             exc,
                         )
                         invalid += 1
                 else:
                     proxy_reader = self._missing_proxy_reader(file_obj, selected_folder)
                     self._save_proxy_reader(file_obj.id, proxy_reader, apply_runtime=False, emit_job_change=False)
                     changed_file_ids.append(str(file_obj.id or ""))
                     missing += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         status = "Optimize Preview: linked {} item(s), missing {}".format(matched, missing)
         if invalid:
             status += ", invalid {}".format(invalid)
         self._show_status(status, 4000)

     def remove_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         removed = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 file_id = str(file_obj.id or "")
                 self.cancel_job(file_id)
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 if "proxy_reader" not in data:
                     continue
                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 removed += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         if removed:
             self._show_status("Optimize Preview: removed {} item(s)".format(removed), 3000)

     def delete_and_unlink_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return 0

         deleted = 0
         unlinked = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 file_id = str(file_obj.id or "")
                 self.cancel_job(file_id)
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 proxy_reader = data.get("proxy_reader")
                 if not isinstance(proxy_reader, dict):
                     continue

                 proxy_path = absolute_media_path(proxy_reader.get("path"))
                 if proxy_path and os.path.exists(proxy_path):
                     proxy_abs = os.path.abspath(proxy_path)
                     try:
                         os.remove(proxy_abs)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", proxy_abs, exc_info=1)

                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 unlinked += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         status = "Optimize Preview: deleted {}, unlinked {}".format(deleted, unlinked)
         if unlinked:
             self._show_status(status, 4000)
         return deleted

     def has_internal_project_proxy_files(self):
         proxy_root = self._proxy_root()
         if not proxy_root or not os.path.isdir(proxy_root):
             return False
         for _, _, files in os.walk(proxy_root):
             if files:
                 return True
         return False

     def delete_internal_project_proxy_files(self):
         proxy_root = os.path.abspath(str(self._proxy_root() or ""))
         proxy_root_prefix = proxy_root + os.sep if proxy_root else ""
         if not proxy_root_prefix or not os.path.isdir(proxy_root):
             return 0

         deleted = 0
         unlinked = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in File.filter():
                 file_id = str(getattr(file_obj, "id", "") or "")
                 if not file_id:
                     continue
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 proxy_reader = data.get("proxy_reader")
                 if not isinstance(proxy_reader, dict):
                     continue

                 proxy_path = absolute_media_path(proxy_reader.get("path"))
                 if not proxy_path:
                     continue
                 proxy_abs = os.path.abspath(proxy_path)
                 if not proxy_abs.startswith(proxy_root_prefix):
                     continue

                 self.cancel_job(file_id)
                 if os.path.exists(proxy_abs):
                     try:
                         os.remove(proxy_abs)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", proxy_abs, exc_info=1)

                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 unlinked += 1
         finally:
             updates.transaction_id = previous_tid

         try:
             for root, dirs, files in os.walk(proxy_root, topdown=False):
                 for name in files:
                     file_path = os.path.join(root, name)
                     try:
                         os.remove(file_path)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", file_path, exc_info=1)
                 for name in dirs:
                     dir_path = os.path.join(root, name)
                     try:
                         os.rmdir(dir_path)
                     except OSError:
                         pass
         except Exception:
             log.warning("Optimize Preview cleanup failed for %s", proxy_root, exc_info=1)

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         if deleted or unlinked:
             self._show_status("Optimize Preview: deleted {}, unlinked {}".format(deleted, unlinked), 4000)
         return deleted

     def cancel_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return False

         canceled = 0
         for file_obj in files:
             if self.cancel_job(file_obj.id):
                 canceled += 1

         if canceled:
             self._show_status("Optimize Preview: canceled {} item(s)".format(canceled), 3000)
             return True
         return False

     def cancel_job(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return False

         with self._lock:
             job = self._jobs.get(file_id)
             if not job:
                 return False
             status = str(job.get("status") or "")
             future = job.get("future")
             if status == "queued" and future and future.cancel():
                 self._finalize_job(file_id, "canceled")
                 return True
             if status == "queued":
                 job["cancel_requested"] = True
                 job["status"] = "canceling"
             elif status == "running":
                 job["cancel_requested"] = True
                 job["status"] = "canceling"
             elif status == "canceling":
                 return True
             else:
                 return False

         self._emit_job_change(file_id)
         return True

     def rewrite_json_for_preview(self, payload):
         """Return preview-bound JSON/object with proxy readers substituted when available."""
         original_is_text = isinstance(payload, str)
         if original_is_text:
             try:
                 parsed = json.loads(payload)
             except Exception:
                 return payload
         else:
             parsed = copy.deepcopy(payload)

         proxy_lookup, path_lookup = self._proxy_lookup_from_payload(parsed)
         if not proxy_lookup:
             proxy_lookup, path_lookup = self._proxy_lookup_from_project()
         if not proxy_lookup:
             return payload

         changed = self._rewrite_payload_in_place(parsed, proxy_lookup, path_lookup)
         if not changed:
             return payload
         if original_is_text:
             return json.dumps(parsed)
         return parsed

     def apply_runtime_updates_for_file(self, file_id):
         """Push runtime-only clip/effect updates to libopenshot for one file's proxy state."""
         return self.apply_runtime_updates_for_files([file_id])

     def apply_runtime_updates_for_files(self, file_ids):
         """Push one combined runtime-only clip/effect diff for one or more file proxy changes."""
         normalized_ids = {str(file_id or "") for file_id in (file_ids or []) if str(file_id or "")}
         if not normalized_ids:
             return False

         timeline_sync = getattr(self.win, "timeline_sync", None)
         timeline = getattr(timeline_sync, "timeline", None) if timeline_sync else None
         if timeline is None:
             return False

         actions = []
         for clip in Clip.filter():
             if any(self._payload_references_file(clip.data, file_id) for file_id in normalized_ids):
                 actions.append(
                     UpdateAction(
                         type="update",
                         key=["clips", {"id": clip.id}],
                         values=copy.deepcopy(clip.data),
                     )
                 )
         for effect in Transition.filter():
             if any(self._payload_references_file(effect.data, file_id) for file_id in normalized_ids):
                 actions.append(
                     UpdateAction(
                         type="update",
                         key=["effects", {"id": effect.id}],
                         values=copy.deepcopy(effect.data),
                     )
                 )

         if not actions:
             return False

         payload = [json.loads(action.json()) for action in actions]
         timeline.ApplyJsonDiff(self.rewrite_json_for_preview(json.dumps(payload)))
         try:
             timeline.ClearAllCache(True)
         except Exception:
             log.debug("Optimize Preview cache clear failed", exc_info=1)
         if getattr(self.win, "refreshFrameSignal", None):
             self.win.refreshFrameSignal.emit()
         return True

     def has_proxy_reader(self, file_obj):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         proxy_reader = data.get("proxy_reader") if isinstance(data, dict) else None
         return isinstance(proxy_reader, dict) and bool(proxy_reader.get("path"))

     def has_missing_proxy(self, file_obj):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         proxy_reader = data.get("proxy_reader") if isinstance(data, dict) else None
         if not isinstance(proxy_reader, dict) or not proxy_reader.get("path"):
             return False
         if proxy_reader.get("missing"):
             return True
         return not os.path.exists(str(proxy_reader.get("path")))

     def get_active_job_for_file(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return None
         with self._lock:
             job = self._job_snapshot(self._jobs.get(file_id))
         if not job or job.get("status") not in self.ACTIVE_STATES:
             return None
         return job

     def get_file_badge(self, file_id):
         job = self.get_active_job_for_file(file_id)
         if not job:
             return None

         status = str(job.get("status") or "").strip().lower()
         progress = int(job.get("progress", 0))
         if status == "queued":
             label = "Queued"
         elif status == "running":
             label = "Creating {}%".format(progress)
         elif status == "canceling":
             label = "Canceling..."
         else:
             label = status.capitalize()
         return {
             "status": status,
             "progress": progress,
             "label": label,
             "job_id": str(job.get("id") or ""),
         }

     def get_proxy_state(self, file_obj):
         file_id = str(getattr(file_obj, "id", "") or "")
         active_job = self.get_active_job_for_file(file_id)
         if active_job:
             return str(active_job.get("status") or "queued")
         if self.has_proxy_reader(file_obj):
             return "missing" if self.has_missing_proxy(file_obj) else "ready"
         return "none"

     def _proxy_root(self):
         app = get_app()
         project = getattr(app, "project", None) if app else None
         current_filepath = getattr(project, "current_filepath", None) if project else None
         if current_filepath:
             current_abs = os.path.abspath(str(current_filepath))
             backup_abs = os.path.abspath(info.BACKUP_FILE)
             recovery_abs = os.path.abspath(info.RECOVERY_PATH) + os.sep
             if current_abs == backup_abs or current_abs.startswith(recovery_abs):
                 return info.PROXY_PATH
             return os.path.join(get_assets_path(current_filepath), "optimized")
         return info.PROXY_PATH

     def _emit_proxy_result(self, file_id, future):
         error_text = ""
         proxy_reader = {}
         try:
             if future.cancelled():
                 raise _ProxyJobCanceled()
             proxy_reader = future.result()
         except _ProxyJobCanceled:
             error_text = "__canceled__"
         except Exception as ex:
             error_text = str(ex)
             log.warning("Optimized Preview generation failed for %s: %s", file_id, error_text, exc_info=1)
         self.proxy_generated.emit(str(file_id or ""), proxy_reader, error_text)

     @pyqtSlot(str, object, str)
     def _on_proxy_generated(self, file_id, proxy_reader, error_text):
         file_id = str(file_id or "")
         if error_text == "__canceled__":
             self._finalize_job(file_id, "canceled")
             return
         if error_text:
             self._finalize_job(file_id, "failed")
             self._show_status("Optimize Preview: {}".format(error_text), 5000)
             return
         self._save_proxy_reader(file_id, proxy_reader)
         self._finalize_job(file_id, "completed")
         self._show_status("Optimize Preview: ready", 3000)

     def _save_proxy_reader(self, file_id, proxy_reader, apply_runtime=True, emit_job_change=True):
         file_obj = File.get(id=file_id)
         if not file_obj:
             return
         file_obj.data["proxy_reader"] = copy.deepcopy(proxy_reader or {})
         file_obj.save()
         if apply_runtime:
             self.apply_runtime_updates_for_file(file_id)
         if emit_job_change:
             self._emit_job_change(file_id)

     def _reader_json_for_path(self, file_path, file_id):
         clip = openshot.Clip(str(file_path))
         try:
             clip.Open()
             proxy_reader = json.loads(clip.Reader().Json())
             proxy_reader["id"] = str(file_id or "")
             return proxy_reader
         finally:
             try:
                 clip.Close()
             except Exception:
                 pass

     def _build_proxy_reader(self, file_id, file_data):
         source_path = absolute_media_path(file_data.get("path"))
         if not source_path or not os.path.exists(source_path):
             raise RuntimeError("source file not found")
         if str(file_data.get("media_type", "")).lower() != "video":
             raise RuntimeError("only video files are supported in V1")

         self._mark_running(file_id)
         output_path = self._reserved_or_computed_proxy_output_path(file_id, file_data)
         try:
             max_width, max_height = self._max_optimize_bounds()
             optimize_timeline = self._create_optimize_timeline(max_width, max_height)
             clip = openshot.Clip(source_path)
             clip.ParentTimeline(optimize_timeline)
             clip.Open()
             try:
                 reader = clip.Reader()
                 self._configure_optimize_caches(clip, reader)
                 source_reader = json.loads(reader.Json())

                 width = int(source_reader.get("width", 0) or 0)
                 height = int(source_reader.get("height", 0) or 0)
                 if width <= 0 or height <= 0:
                     raise RuntimeError("invalid source dimensions")

                 target_width, target_height = self._scaled_dimensions(width, height, max_width, max_height)
                 fps = source_reader.get("fps", {"num": 30, "den": 1})
                 pixel_ratio = source_reader.get("pixel_ratio", {"num": 1, "den": 1})
                 writer = openshot.FFmpegWriter(output_path)
                 writer.SetVideoOptions(
                     True,
                     "libx264",
                     openshot.Fraction(int(fps.get("num", 30)), int(fps.get("den", 1))),
                     target_width,
                     target_height,
                     openshot.Fraction(int(pixel_ratio.get("num", 1)), int(pixel_ratio.get("den", 1))),
                     False,
                     False,
                     28,
                 )
                 writer.PrepareStreams()

                 if source_reader.get("has_audio"):
                     channel_layout = openshot.LAYOUT_STEREO if int(source_reader.get("channels", 2) or 2) > 1 else openshot.LAYOUT_MONO
                     writer.SetAudioOptions(
                         True,
                         "aac",
                         48000,
                         2 if channel_layout == openshot.LAYOUT_STEREO else 1,
                         channel_layout,
                         128000,
                     )
                     writer.PrepareStreams()

                 writer.Open()
                 try:
                     max_frame = int(source_reader.get("video_length", 0) or 0)
                     if max_frame <= 0:
                         raise RuntimeError("invalid source frame count")
                     rotate = self._rotation_for_reader(reader)
                     prewarm_frames = self._thumbnail_prewarm_frames(
                         file_id,
                         max_frame,
                         fps,
                         thumbs_per_second=self._thumbnail_prewarm_rate(),
                     )
                     prewarm_index = 0
                     for frame_number in range(1, max_frame + 1):
                         self._raise_if_canceled(file_id)
                         frame = reader.GetFrame(frame_number)
                         writer.WriteFrame(frame)
                         while prewarm_index < len(prewarm_frames) and prewarm_frames[prewarm_index] < frame_number:
                             prewarm_index += 1
                         if prewarm_index < len(prewarm_frames) and prewarm_frames[prewarm_index] == frame_number:
                             self._save_prewarmed_thumbnail(frame, file_id, frame_number, rotate)
                             prewarm_index += 1
                         self._trim_optimize_caches(clip, reader, frame_number)
                         if frame_number == 1 or frame_number == max_frame or frame_number % max(1, min(12, max_frame // 40 or 1)) == 0:
                             self._update_progress(file_id, int((float(frame_number) / float(max_frame)) * 100.0))
                 finally:
                     writer.Close()
                     self._clear_cache_object(getattr(reader, "GetCache", lambda: None)())
                     self._clear_cache_object(getattr(clip, "GetCache", lambda: None)())
             finally:
                 clip.Close()
                 clip.ParentTimeline(None)
         except Exception:
             try:
                 if os.path.exists(output_path):
                     os.remove(output_path)
             except Exception:
                 pass
             raise

         return self._reader_json_for_path(output_path, file_id)

     def _proxy_lookup_from_payload(self, payload):
         if isinstance(payload, dict):
             files = payload.get("files")
             if isinstance(files, list):
                 return self._proxy_lookup_from_files(files)
         return {}, {}

     def _proxy_lookup_from_project(self):
         app = get_app()
         project = getattr(app, "project", None) if app else None
         files = project.get("files") if project else None
         if not isinstance(files, list):
             return {}, {}
         return self._proxy_lookup_from_files(files)

     def _proxy_lookup_from_files(self, files):
         proxy_lookup = {}
         path_lookup = {}
         for file_data in files or []:
             if not isinstance(file_data, dict):
                 continue
             file_id = str(file_data.get("id") or "")
             file_path = file_data.get("path")
             if file_id and file_path:
                 path_lookup[comparable_media_path(file_path)] = file_id
             proxy_reader = file_data.get("proxy_reader")
             if not isinstance(proxy_reader, dict) or not proxy_reader.get("path") or not file_id:
                 continue
             if not os.path.exists(str(proxy_reader.get("path"))):
                 continue
             proxy_data = copy.deepcopy(proxy_reader)
             proxy_data["id"] = file_id
             proxy_lookup[file_id] = proxy_data
         return proxy_lookup, path_lookup

     def _rewrite_payload_in_place(self, payload, proxy_lookup, path_lookup):
         changed = False
         if isinstance(payload, dict):
             for key, value in list(payload.items()):
                 if key in ("reader", "mask_reader") and isinstance(value, dict):
                     replacement = self._proxy_replacement_for_reader(value, proxy_lookup, path_lookup)
                     if replacement is not None and replacement != value:
                         payload[key] = copy.deepcopy(replacement)
                         changed = True
                 else:
                     changed = self._rewrite_payload_in_place(value, proxy_lookup, path_lookup) or changed
         elif isinstance(payload, list):
             for item in payload:
                 changed = self._rewrite_payload_in_place(item, proxy_lookup, path_lookup) or changed
         return changed

     def _proxy_replacement_for_reader(self, reader_data, proxy_lookup, path_lookup):
         reader_id = str(reader_data.get("id") or "")
         if reader_id and reader_id in proxy_lookup:
             return proxy_lookup[reader_id]

         reader_path = reader_data.get("path")
         if reader_path:
             file_id = path_lookup.get(comparable_media_path(reader_path))
             if file_id and file_id in proxy_lookup:
                 return proxy_lookup[file_id]
         return None

     def _payload_references_file(self, payload, file_id):
         file_obj = File.get(id=file_id)
         if not file_obj:
             return False
         target_path = comparable_media_path((file_obj.data or {}).get("path"))
         return self._payload_contains_file_reference(payload, str(file_id), target_path)

     def _payload_contains_file_reference(self, payload, file_id, target_path):
         if isinstance(payload, dict):
             for key, value in payload.items():
                 if key in ("reader", "mask_reader") and isinstance(value, dict):
                     if str(value.get("id") or "") == file_id:
                         return True
                     reader_path = value.get("path")
                     if reader_path and comparable_media_path(reader_path) == target_path:
                         return True
                 if self._payload_contains_file_reference(value, file_id, target_path):
                     return True
         elif isinstance(payload, list):
             for item in payload:
                 if self._payload_contains_file_reference(item, file_id, target_path):
                     return True
         return False

     def _show_status(self, text, timeout=5000):
         status_bar = getattr(self.win, "statusBar", None)
         if status_bar and hasattr(status_bar, "showMessage"):
             status_bar.showMessage(str(text), int(timeout))

     def _index_existing_optimized_files(self, folder_path):
         basename_index = {}
         stem_index = {}
         normalized_index = {}
         path_index = {}
         for root, _, files in os.walk(str(folder_path or "")):
             for name in files:
                 ext = os.path.splitext(name)[1].lower()
                 if ext not in self.MATCHABLE_VIDEO_EXTENSIONS:
                     continue
                 full_path = os.path.join(root, name)
                 basename_index.setdefault(name.lower(), []).append(full_path)
                 stem = os.path.splitext(name)[0].lower()
                 stem_index.setdefault(stem, []).append(full_path)
                 normalized_index.setdefault(self._normalized_proxy_name_key(stem), []).append(full_path)
                 path_index[comparable_media_path(full_path)] = full_path
         return {
             "basename": basename_index,
             "stem": stem_index,
             "normalized": normalized_index,
             "path": path_index,
         }

     def _match_existing_optimized_path(self, file_obj, folder_path, folder_index):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         source_path = absolute_media_path(data.get("path"))
         file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
         basename = os.path.basename(source_path or "")
         basename_lower = basename.lower()
         source_stem = os.path.splitext(basename_lower)[0]
         source_norm = comparable_media_path(source_path)

         if file_id:
             id_matches = folder_index.get("stem", {}).get(file_id.lower(), [])
             if id_matches:
                 return id_matches[0]

         if basename_lower:
             basename_matches = folder_index.get("basename", {}).get(basename_lower, [])
             if basename_matches:
                 return basename_matches[0]

         if source_stem:
             stem_matches = folder_index.get("stem", {}).get(source_stem, [])
             if stem_matches:
                 return stem_matches[0]

         source_variants = self._source_proxy_name_candidates(source_stem)
         candidate_scores = {}
         for variant in source_variants:
             for match_path in folder_index.get("stem", {}).get(variant, []):
                 candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 90)

         normalized_key = self._normalized_proxy_name_key(source_stem)
         if normalized_key:
             for match_path in folder_index.get("normalized", {}).get(normalized_key, []):
                 candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 80)

         source_prefixes = self._source_stem_file_id_prefix_candidates(source_stem, file_id)
         for candidate_stem, candidate_paths in folder_index.get("stem", {}).items():
             if source_prefixes and any(candidate_stem.startswith(prefix) for prefix in source_prefixes):
                 for match_path in candidate_paths:
                     candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 75)

         for candidate_path in folder_index.get("path", {}).values():
             if source_norm and comparable_media_path(candidate_path).endswith(source_norm):
                 candidate_scores[candidate_path] = max(candidate_scores.get(candidate_path, 0), 70)

         if candidate_scores:
             ranked_paths = sorted(
                 candidate_scores.items(),
                 key=lambda item: (
                     -item[1],
                     abs(len(os.path.splitext(os.path.basename(item[0]))[0]) - len(source_stem)),
                     os.path.basename(item[0]).lower(),
                 ),
             )
             return ranked_paths[0][0]

         expected_name = self._preferred_proxy_filename(file_id, data, folder_path, existing_names=set())
         return os.path.join(str(folder_path or ""), expected_name)

     def _missing_proxy_reader(self, file_obj, folder_path):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
         source_path = absolute_media_path(data.get("path"))
         basename = os.path.basename(source_path or "")
         expected_path = self._match_existing_optimized_path(file_obj, folder_path, {"basename": {}, "stem": {}, "path": {}})
         return {
             "id": file_id,
             "path": expected_path,
             "missing": True,
             "name": basename,
         }

     def _proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         os.makedirs(proxy_root, exist_ok=True)
         filename = self._preferred_proxy_filename(file_id, file_data, proxy_root)
         return os.path.join(proxy_root, filename)

     def _existing_proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         default_filename = self._preferred_proxy_filename(file_id, file_data, proxy_root, existing_names=set())
         default_path = os.path.join(proxy_root, default_filename)
         source_stem = os.path.splitext(os.path.basename(absolute_media_path((file_data or {}).get("path")) or ""))[0]
         specific_path = os.path.join(proxy_root, "{}_{}.mp4".format(self._proxy_filename_stem(source_stem), str(file_id or "")))

         if os.path.exists(specific_path):
             return specific_path
         if os.path.exists(default_path):
             return default_path
         return None

     def _reserve_proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         os.makedirs(proxy_root, exist_ok=True)
         try:
             existing_names = set(os.listdir(proxy_root))
         except Exception:
             existing_names = set()
         with self._lock:
             for job in self._jobs.values():
                 output_path = job.get("output_path")
                 if output_path:
                     existing_names.add(os.path.basename(str(output_path)))
         filename = self._preferred_proxy_filename(file_id, file_data, proxy_root, existing_names=existing_names)
         return os.path.join(proxy_root, filename)

     def _reserved_or_computed_proxy_output_path(self, file_id, file_data):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             reserved_path = job.get("output_path") if isinstance(job, dict) else None
         return reserved_path or self._proxy_output_path(file_id, file_data)

     def _preferred_proxy_filename(self, file_id, file_data, proxy_root, existing_names=None):
         file_id = str(file_id or "")
         data = file_data or {}
         source_path = absolute_media_path(data.get("path"))
         source_stem = os.path.splitext(os.path.basename(source_path or ""))[0]
         base_stem = self._proxy_filename_stem(source_stem)
         default_name = "{}.mp4".format(base_stem)

         if existing_names is None:
             try:
                 existing_names = set(os.listdir(proxy_root))
             except Exception:
                 existing_names = set()
         existing_names_lower = {str(name).lower() for name in existing_names}

         if default_name.lower() not in existing_names_lower:
             return default_name
         return "{}_{}.mp4".format(base_stem, file_id)

     @staticmethod
     def _proxy_filename_stem(source_stem):
         source_stem = str(source_stem or "").strip()
         if not source_stem:
             return "optimized_proxy"
         return "{}_proxy".format(source_stem)


     @staticmethod
     def _source_stem_file_id_prefix_candidates(source_stem, file_id):
         source_stem = str(source_stem or "").strip().lower()
         file_id = str(file_id or "").strip().lower()
         if not source_stem or not file_id:
             return []
         separators = ("_", "-", ".")
         prefixes = []
         for separator in separators:
             prefixes.append("{}{}{}".format(source_stem, separator, file_id))
             prefixes.append("{}{}proxy{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}preview{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}optimized{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}lowres{}{}".format(source_stem, separator, separator, file_id))
         return prefixes

     @classmethod
     def _normalized_proxy_name_key(cls, stem):
         normalized = str(stem or "").strip().lower()
         if not normalized:
             return ""
         normalized = re.sub(r"[\s.-]+", "_", normalized)
         normalized = re.sub(r"_+", "_", normalized).strip("_")
         previous = None
         while normalized and normalized != previous:
             previous = normalized
             normalized = cls._PROXY_PREFIX_RE.sub("", normalized).strip("._- ")
             normalized = cls._PROXY_SUFFIX_RE.sub("", normalized).strip("._- ")
             normalized = cls._RESOLUTION_SUFFIX_RE.sub("", normalized).strip("._- ")
             if cls._CAMERA_PROXY_SUFFIX_RE.search(normalized):
                 trimmed = cls._CAMERA_PROXY_SUFFIX_RE.sub("", normalized).strip("._- ")
                 if trimmed:
                     normalized = trimmed
         return normalized

     @classmethod
     def _source_proxy_name_candidates(cls, source_stem):
         source_stem = str(source_stem or "").strip().lower()
         if not source_stem:
             return []

         candidates = {
             source_stem,
             "{}_proxy".format(source_stem),
             "{}-proxy".format(source_stem),
             "{}_optimized".format(source_stem),
             "{}-optimized".format(source_stem),
             "{}_preview".format(source_stem),
             "{}-preview".format(source_stem),
             "{}_lowres".format(source_stem),
             "{}-lowres".format(source_stem),
             "{}_p".format(source_stem),
             "{}-p".format(source_stem),
             "proxy_{}".format(source_stem),
             "proxy-{}".format(source_stem),
             "optimized_{}".format(source_stem),
             "optimized-{}".format(source_stem),
             "preview_{}".format(source_stem),
             "preview-{}".format(source_stem),
         }
         for size_label in ("360p", "540p", "720p", "1080p", "1440p", "2160p", "4k"):
             candidates.add("{}_{}".format(source_stem, size_label))
             candidates.add("{}_proxy_{}".format(source_stem, size_label))
             candidates.add("{}_{}_proxy".format(source_stem, size_label))
         return sorted(candidates)

     def _thumbnail_prewarm_frames(self, file_id, max_frame, fps, thumbs_per_second=None):
         max_frame = max(1, int(max_frame or 1))
         fps_num = float((fps or {}).get("num", 0.0) or 0.0)
         fps_den = float((fps or {}).get("den", 1.0) or 1.0)
         fps_value = (fps_num / fps_den) if fps_num > 0.0 and fps_den > 0.0 else 0.0
         if fps_value <= 0.0:
             return [1]

         if thumbs_per_second is None:
             thumbs_per_second = self._thumbnail_prewarm_rate()
         thumbs_per_second = max(1, int(thumbs_per_second))
         step = max(1, int(round(fps_value / float(thumbs_per_second))))
         frames = []
         seen = set()
         for frame_number in range(1, max_frame + 1, step):
             rounded = min(max_frame, RoundFrameToThumbnailGrid(frame_number, fps_value))
             if rounded not in seen:
                 seen.add(rounded)
                 frames.append(rounded)
         if max_frame not in seen:
             frames.append(max_frame)
         if 1 not in seen:
             frames.insert(0, 1)
         return sorted(set(frames))

     def _save_prewarmed_thumbnail(self, frame, file_id, frame_number, rotate):
         thumb_path = ThumbnailPathForFrame(file_id, frame_number)
         if os.path.exists(thumb_path):
             return
         try:
             mask_path = os.path.join(info.IMAGES_PATH, "mask.png")
             GenerateThumbnailFromFrame(frame, thumb_path, 98, 64, mask_path, "", rotate=rotate)
         except Exception:
             log.debug("Optimize Preview thumbnail prewarm failed file=%s frame=%s", file_id, frame_number, exc_info=1)

     def _configure_optimize_caches(self, clip, reader):
         self._configure_cache_object(getattr(clip, "GetCache", lambda: None)())
         self._configure_cache_object(getattr(reader, "GetCache", lambda: None)())

     def _trim_optimize_caches(self, clip, reader, frame_number):
         if int(frame_number or 0) % int(self.OPTIMIZE_CACHE_CLEAR_INTERVAL) != 0:
             return
         self._clear_cache_object(getattr(reader, "GetCache", lambda: None)())
         self._clear_cache_object(getattr(clip, "GetCache", lambda: None)())

     def _configure_cache_object(self, cache_object):
         if not cache_object:
             return
         try:
             cache_object.SetMaxBytes(int(self.OPTIMIZE_CACHE_MAX_BYTES))
         except Exception:
             log.debug("Optimize Preview cache max-bytes update failed", exc_info=1)
         self._clear_cache_object(cache_object)

     @staticmethod
     def _clear_cache_object(cache_object):
         if not cache_object:
             return
         try:
             cache_object.Clear()
         except Exception:
             log.debug("Optimize Preview cache clear failed", exc_info=1)

     @staticmethod
     def _rotation_for_reader(reader):
         try:
             if reader.info.metadata.count("rotate"):
                 return float(reader.info.metadata["rotate"])
         except Exception:
             pass
         return 0.0

     def _create_optimize_timeline(self, width, height):
         width = max(2, int(width or 2))
         height = max(2, int(height or 2))
         timeline = openshot.Timeline(
             width,
             height,
             openshot.Fraction(30, 1),
             48000,
             2,
             openshot.LAYOUT_STEREO,
         )
         timeline.preview_width = width
         timeline.preview_height = height
         return timeline

     def _ensure_executor(self):
         worker_count = self._optimize_worker_count()
         current_executor = getattr(self, "_executor", None)
         current_count = getattr(current_executor, "_max_workers", None)
         if current_executor and current_count == worker_count:
             return
         with self._lock:
             if self._jobs:
                 return
             old_executor = self._executor
             self._executor = ThreadPoolExecutor(
                 max_workers=worker_count,
                 thread_name_prefix="optimized-preview",
             )
         if old_executor:
             self._shutdown_executor(old_executor)

     @staticmethod
     def _shutdown_executor(executor):
         if not executor:
             return
         try:
             executor.shutdown(wait=False, cancel_futures=True)
         except TypeError:
             executor.shutdown(wait=False)

     def _optimize_worker_count(self):
         value = self._setting_value("optimize-preview-jobs", 1)
         try:
             return max(1, min(8, int(value)))
         except (TypeError, ValueError):
             return 1

     def _thumbnail_prewarm_rate(self):
         value = self._setting_value("optimize-preview-thumbnails", 4)
         try:
             return max(1, min(12, int(value)))
         except (TypeError, ValueError):
             return 4

     def _max_optimize_bounds(self):
         raw_value = str(self._setting_value("optimize-preview-max-size", "1280x720") or "1280x720").lower()
         try:
             width_text, height_text = raw_value.split("x", 1)
             width = max(2, int(width_text))
             height = max(2, int(height_text))
             return width, height
         except (AttributeError, TypeError, ValueError):
             return 1280, 720

     @staticmethod
     def _setting_value(name, default):
         app = get_app()
         settings = app.get_settings() if app else None
         if not settings:
             return default
         try:
             value = settings.get(name)
         except Exception:
             return default
         return default if value is None else value

     def _mark_running(self, file_id):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if not job:
                 return
             if job.get("cancel_requested"):
                 job["status"] = "canceling"
             else:
                 job["status"] = "running"
             job["progress"] = max(1, int(job.get("progress", 0)))
         self._emit_job_change(file_id)

     def _update_progress(self, file_id, progress):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if not job:
                 return
             if job.get("status") not in self.ACTIVE_STATES:
                 return
             job["progress"] = max(0, min(100, int(progress)))
         self._emit_job_change(file_id)

     def _raise_if_canceled(self, file_id):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if job and job.get("cancel_requested"):
                 raise _ProxyJobCanceled()

     def _finalize_job(self, file_id, status):
         file_id = str(file_id or "")
         with self._lock:
             job = self._jobs.pop(file_id, None)
         if status == "completed":
             self.job_finished.emit(file_id, status)
         elif job is not None:
             self.job_finished.emit(file_id, status)
         self._emit_job_change(file_id)

     def _emit_job_change(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return
         job = self.get_active_job_for_file(file_id)
         if job:
             self.job_updated.emit(file_id, str(job.get("status") or ""), int(job.get("progress", 0)))
         self.file_job_changed.emit(file_id)
         self.queue_changed.emit()

     @staticmethod
     def _job_snapshot(job):
         if not isinstance(job, dict):
             return None
         return {
             "id": str(job.get("id") or ""),
             "status": str(job.get("status") or ""),
             "progress": int(job.get("progress", 0)),
             "cancel_requested": bool(job.get("cancel_requested")),
         }

     @staticmethod
     def _scaled_dimensions(width, height, max_width=1280, max_height=720):
         if width <= 0 or height <= 0:
             return (1280, 720)
         scale = min(float(max_width) / float(width), float(max_height) / float(height), 1.0)
         scaled_w = max(2, int(round(width * scale)))
         scaled_h = max(2, int(round(height * scale)))
         if scaled_w % 2:
             scaled_w -= 1
         if scaled_h % 2:
             scaled_h -= 1
         return (max(2, scaled_w), max(2, scaled_h))


class _ProxyJobCanceled(RuntimeError):
     """Raised when a queued/running proxy job is canceled."""



"""
@file
@brief This file is used to import an EDL (edit decision list) file
@author Jonathan Thomas <jonathan@openshot.org>

@section LICENSE

Copyright (c) 2008-2018 OpenShot Studios, LLC
(http://www.openshotstudios.com). This file is part of
OpenShot Video Editor (http://www.openshot.org), an open-source project
dedicated to delivering high quality video editing and animation solutions
to the world.

OpenShot Video Editor is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenShot Video Editor is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os
import re
from operator import itemgetter

import openshot
from PyQt5.QtWidgets import QFileDialog

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.image_types import get_media_type
from classes.path_utils import absolute_path_from_export
from classes.query import Clip, Track, File
from classes.time_parts import timecodeToSeconds
from windows.views.find_file import find_missing_file

# REGEX expressions to parse lines from EDL file
title_regex = re.compile(r"TITLE:[ ]+(.*)")
clips_regex = re.compile(r"(\d{3})[ ]+(.+?)[ ]+(.+?)[ ]+(.+?)[ ]+(.*)[ ]+(.*)[ ]+(.*)[ ]+(.*)")
clip_name_regex = re.compile(r"[*][ ]+FROM CLIP NAME:[ ]+(.*)")
source_regex = re.compile(r"[*][ ]+SOURCE FILE:[ ]+(.*)")
_PERMISSIVE = r"[ ]*;?[ ]*(?:interp[:=])?[ ]*([A-Za-z0-9]+)?(?:.*)?$"
param_regexes = [
    ("opacity", re.compile(r"\* (?:OPACITY|VIDEO) LEVEL AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("volume", re.compile(r"\* AUDIO LEVEL AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*d[bB]?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("scale_x", re.compile(r"\* SCALE X AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("scale_y", re.compile(r"\* SCALE Y AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("location_x", re.compile(r"\* LOCATION X AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("location_y", re.compile(r"\* LOCATION Y AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("rotation", re.compile(r"\* ROTATION AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*(?:deg)?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("shear_x", re.compile(r"\* SHEAR X AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
    ("shear_y", re.compile(r"\* SHEAR Y AT\s+([0-9:;]+)\s+IS\s+([+-]?[0-9.]+)\s*%?\s*" + _PERMISSIVE, re.IGNORECASE)),
]
fcm_regex = re.compile(r"FCM:[ ]+(.*)")


def _interp_from_name(name):
    n = (str(name) if name is not None else "").strip().lower()
    # libopenshot: 0=bezier, 1=linear, 2=constant
    if n.isdigit():
        try:
            numeric = int(n)
        except ValueError:
            numeric = None
    else:
        numeric = None
    if numeric == 0:
        return openshot.BEZIER
    if numeric == 1:
        return openshot.LINEAR
    if numeric == 2:
        return openshot.CONSTANT
    if n.startswith("bez"):
        return openshot.BEZIER
    if n.startswith("hold") or n.startswith("const"):
        return openshot.CONSTANT
    return openshot.LINEAR


def _db_to_volume(db_value):
    """Convert dB to linear 0-1, with floor at -96 dB."""
    try:
        db = float(db_value)
    except (TypeError, ValueError):
        return 0.0
    if db <= -96.0:
        return 0.0
    linear = 10 ** (db / 20.0)
    return max(0.0, min(1.0, linear))


def create_clip(context, track):
    """Create a new clip based on this context dict"""
    app = get_app()
    _ = app._tr

    # Get FPS info
    fps_num = app.project.get("fps").get("num", 24)
    fps_den = app.project.get("fps").get("den", 1)
    fps_float = float(fps_num / fps_den)

    clip_path_value = context.get("clip_path") or context.get("clip_title") or ""
    clip_path_value = clip_path_value or ""

    # Get clip path (and prompt user if path not found)
    clip_path, is_modified, is_skipped = find_missing_file(clip_path_value)
    if is_skipped:
        return

    # Get component contexts
    video_ctx = context.get("video_ctx", {})
    audio_ctx_list = context.get("audio_ctx", [])
    audio_ctx = audio_ctx_list[0] if audio_ctx_list else {}

    if not (video_ctx or audio_ctx):
        # Nothing to import (likely filler)
        return

    # Check for this path in our existing project data
    file = File.get(path=clip_path)

    # Load filepath in libopenshot clip object (which will try multiple readers to open it)
    clip_obj = openshot.Clip(clip_path)

    if not file:
        # Get the JSON for the clip's internal reader
        try:
            reader = clip_obj.Reader()
            file_data = json.loads(reader.Json())

            # Determine media type
            file_data["media_type"] = get_media_type(file_data)

            # Save new file to the project data
            file = File()
            file.data = file_data

            # Save file
            file.save()
        except Exception:
            log.warning("Error building File object for %s" % clip_path, exc_info=1)

    if file.data["media_type"] == "video" or file.data["media_type"] == "image":
        # Determine thumb path
        thumb_path = os.path.join(info.THUMBNAIL_PATH, "%s.png" % file.data["id"])
    else:
        # Audio file
        thumb_path = os.path.join(info.PATH, "images", "AudioThumbnail.png")

    # Create Clip object
    clip = Clip()
    clip.data = json.loads(clip_obj.Json())
    clip.data["file_id"] = file.id
    clip_title = context.get("clip_title") or os.path.basename(clip_path_value) or clip_path_value
    clip.data["title"] = clip_title
    clip.data["layer"] = track.data.get("number", 1000000)
    reel_name = (video_ctx or audio_ctx).get("reel") if (video_ctx or audio_ctx) else None
    if not reel_name and audio_ctx_list:
        reel_name = audio_ctx_list[0].get("reel")
    if reel_name:
        clip.data["reel"] = reel_name

    if video_ctx and not audio_ctx:
        # Only video
        clip.data["position"] = timecodeToSeconds(
            video_ctx.get("timeline_position", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["start"] = timecodeToSeconds(
            video_ctx.get("clip_start_time", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["end"] = timecodeToSeconds(
            video_ctx.get("clip_end_time", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["has_audio"] = {
            "Points": [
                {
                    "co": {
                        "X": 1.0,
                        "Y": 0.0,  # Disable audio
                    },
                    "interpolation": 2,
                }
            ]
        }
    elif audio_ctx and not video_ctx:
        # Only audio
        clip.data["position"] = timecodeToSeconds(
            audio_ctx.get("timeline_position", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["start"] = timecodeToSeconds(
            audio_ctx.get("clip_start_time", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["end"] = timecodeToSeconds(
            audio_ctx.get("clip_end_time", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["has_video"] = {
            "Points": [
                {
                    "co": {
                        "X": 1.0,
                        "Y": 0.0,  # Disable video
                    },
                    "interpolation": 2,
                }
            ]
        }
    else:
        # Both video and audio
        clip.data["position"] = timecodeToSeconds(
            video_ctx.get("timeline_position", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["start"] = timecodeToSeconds(
            video_ctx.get("clip_start_time", "00:00:00:00"), fps_num, fps_den
        )
        clip.data["end"] = timecodeToSeconds(
            video_ctx.get("clip_end_time", "00:00:00:00"), fps_num, fps_den
        )

    # Add volume keyframes
    if context.get("volume"):
        clip.data["volume"] = {"Points": []}
        for keyframe in context.get("volume", []):
            clip.data["volume"]["Points"].append(
                {
                    "co": {
                        "X": round(
                            timecodeToSeconds(
                                keyframe.get("time", 0.0), fps_num, fps_den
                            ) * fps_float
                        ),
                        "Y": keyframe.get("value", 0.0),
                    },
                    "interpolation": _interp_from_name(keyframe.get("interp")),
                }
            )

    # Add alpha keyframes (from opacity)
    if context.get("opacity"):
        clip.data["alpha"] = {"Points": []}
        for keyframe in context.get("opacity", []):
            clip.data["alpha"]["Points"].append(
                {
                    "co": {
                        "X": round(
                            timecodeToSeconds(
                                keyframe.get("time", 0.0), fps_num, fps_den
                            ) * fps_float
                        ),
                        "Y": keyframe.get("value", 0.0),
                    },
                    "interpolation": _interp_from_name(keyframe.get("interp")),
                }
            )

    # Add transform keyframes
    for field in ("scale_x", "scale_y", "location_x", "location_y", "rotation", "shear_x", "shear_y"):
        if context.get(field):
            clip.data[field] = {"Points": []}
            for keyframe in context.get(field, []):
                clip.data[field]["Points"].append(
                    {
                        "co": {
                            "X": round(
                                timecodeToSeconds(
                                    keyframe.get("time", 0.0), fps_num, fps_den
                                ) * fps_float
                            ),
                            "Y": keyframe.get("value", 0.0),
                        },
                        "interpolation": _interp_from_name(keyframe.get("interp")),
                    }
                )

    # Save clip
    clip.save()


def import_edl():
    """Import EDL File"""
    app = get_app()
    _ = app._tr

    # Get EDL path
    recommended_path = app.project.current_filepath or ""
    if not recommended_path:
        recommended_path = info.HOME_PATH
    else:
        recommended_path = os.path.dirname(recommended_path)
    file_path = QFileDialog.getOpenFileName(
        app.window,
        _("Import EDL..."),
        recommended_path,
        _("Edit Decision List (*.edl)"),
        _("Edit Decision List (*.edl)"),
    )[0]
    if os.path.exists(file_path):
        context = {"audio_ctx": []}
        current_clip_index = ""
        edl_folder = os.path.dirname(os.path.abspath(file_path))

        # Get # of tracks
        all_tracks = app.project.get("layers")
        track_number = list(
            reversed(sorted(all_tracks, key=itemgetter("number")))
        )[0].get("number") + 1000000

        # Create new track above existing layer(s)
        track = Track()
        track.data = {"number": track_number, "y": 0, "label": "EDL Import", "lock": False}
        track.save()

        # Open EDL file
        with open(file_path, "r") as f:
            # Loop through each line, and compare against regex expressions
            for line in f:
                # Detect title
                for r in title_regex.findall(line):
                    context["title"] = r  # Project title

                # Detect clips
                for r in clips_regex.findall(line):
                    if len(r) == 8:
                        edit_index = r[0]   # 001
                        tape = r[1]         # BL, AX
                        clip_type = r[2]    # V, A
                        if tape == "BL":
                            # Ignore
                            continue
                        if current_clip_index == "":
                            # first clip, ignore for now
                            current_clip_index = edit_index
                        if current_clip_index != edit_index:
                            # clip changed, time to commit previous context
                            create_clip(context, track)

                            # reset context
                            current_clip_index = edit_index
                            context = {
                                "title": context.get("title"),
                                "fcm": context.get("fcm"),
                                "audio_ctx": [],
                            }

                        # New clip detected
                        context["edit_index"] = edit_index  # 001

                        component_ctx = {
                            "reel": tape,
                            "edit_type": r[3],
                            "clip_start_time": r[4],
                            "clip_end_time": r[5],
                            "timeline_position": r[6],
                            "timeline_position_end": r[7],
                        }

                        clip_type_key = clip_type.strip().upper()
                        if clip_type_key.startswith("V"):
                            context["video_ctx"] = component_ctx
                        elif clip_type_key.startswith("A"):
                            context.setdefault("audio_ctx", [])
                            context["audio_ctx"].append(component_ctx)

                # Detect clip name
                for r in clip_name_regex.findall(line):
                    context["clip_title"] = r
                    if "clip_path" not in context or not context.get("clip_path"):
                        context["clip_path"] = absolute_path_from_export(r, edl_folder)

                for r in source_regex.findall(line):
                    resolved_path = absolute_path_from_export(r, edl_folder)
                    context["clip_path"] = resolved_path

                # Detect keyframe comments
                for field, regex in param_regexes:
                    for r in regex.findall(line):
                        if len(r) >= 2:
                            context.setdefault(field, [])
                            keyframe_time = r[0]
                            raw_val = r[1]
                            interp_name = r[2].strip() if len(r) > 2 and r[2] else None
                            if not interp_name:
                                m = re.search(r"interp[:=]\s*([^\s)]+)", line, re.IGNORECASE)
                                if m:
                                    interp_name = m.group(1)

                            # NOTE: opacity is stored as 0–1, volume via dB→linear,
                            # and most % based params are normalized 0–1.
                            if field == "opacity":
                                keyframe_value = float(raw_val) / 100.0
                            elif field == "volume":
                                keyframe_value = _db_to_volume(raw_val)
                            elif field in ("scale_x", "scale_y", "location_x", "location_y", "shear_x", "shear_y"):
                                keyframe_value = float(raw_val) / 100.0
                            else:
                                keyframe_value = float(raw_val)

                            context[field].append(
                                {"time": keyframe_time, "value": keyframe_value, "interp": interp_name}
                            )

                # Detect FCM attribute
                for r in fcm_regex.findall(line):
                    context["fcm"] = r   # NON-DROP FRAME

            # Final edit needs committing
            create_clip(context, track)

            # Update the preview and reselect current frame in properties
            app.window.refreshFrameSignal.emit()
            app.window.propertyTableView.select_frame(
                app.window.preview_thread.player.Position()
            )



"""
 @file
 @brief This file contains the classes needed for tracking updates and distributing changes
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from classes.logger import log
from classes.app import get_app
import json
import uuid


class UpdateWatcher:
    """ Interface for classes that listen for 'undo' and 'redo' events. """

    def updateStatusChanged(self, undo_status, redo_status):
        """ Easily be notified each time there are 'undo' or 'redo' actions
        available in the UpdateManager. """
        raise NotImplementedError("updateStatus() not implemented in UpdateWatcher implementer.")


class UpdateInterface:
    """ Interface for classes that listen for changes (insert, update, and delete). """

    def changed(self, action):
        """ This method is invoked each time the UpdateManager is changed.
        The action contains all the details of what changed,
        including the type of change (insert, update, or delete). """
        raise NotImplementedError("changed() not implemented in UpdateInterface implementer.")


class UpdateAction:
    """A data structure representing a single update manager action,
    including any necessary data to reverse the action."""

    def __init__(self, type=None, key=None, values=None, old_values=None, transaction=None):
        self.type = type  # insert, update, or delete
        self.key = key or []  # list which contains the path to the item, for example: ["clips",{"id":"123"}]
        self.values = values
        self.old_values = old_values
        self.transaction = transaction

        # Set transaction id - used with undo/redo to group
        # a set of related UpdateAction objects
        if not transaction:
            # Auto set transaction id to GUID
            self.transaction = str(uuid.uuid4())

    def copy(self):
        """Create and return a copy of UpdateAction - with no references to the original"""
        # Serialize as JSON string, then load JSON string and cast back into UpdateAction
        return UpdateAction(**json.loads(json.dumps(self, default=lambda o: o.__dict__)))

    def set_old_values(self, old_vals):
        self.old_values = old_vals

    def json(self, is_array=False, only_value=False):
        """ Get the JSON string representing this UpdateAction """

        # Build the dictionary to be serialized
        if only_value:
            data_dict = json.loads(json.dumps(self.values))
        else:
            data_dict = {"type": self.type,
                         "key": self.key,
                         "value": json.loads(json.dumps(self.values)),
                         "old_values": json.loads(json.dumps(self.old_values)),
                         "transaction": self.transaction}

            # Always remove 'history' key (if found). This prevents nested "history"
            # attributes when a project dict is loaded.
            try:
                if isinstance(data_dict.get("value"), dict) and "history" in data_dict.get("value"):
                    data_dict.get("value").pop("history", None)
                if isinstance(data_dict.get("old_values"), dict) and "history" in data_dict.get("old_values"):
                    data_dict.get("old_values").pop("history", None)
            except Exception as ex:
                log.warning('Failed to clear history attribute from undo/redo data. {}'.format(ex))

        if not is_array:
            # Use a JSON Object as the root object
            update_action_dict = data_dict
        else:
            # Use a JSON Array as the root object
            update_action_dict = [data_dict]

        # Serialize as JSON
        return json.dumps(update_action_dict)

    def load_json(self, value):
        """ Load this UpdateAction from a JSON string """

        # Load JSON string
        update_action_dict = json.loads(value, strict=False)

        # Set the Update Action properties
        self.type = update_action_dict.get("type")
        self.key = update_action_dict.get("key")
        self.values = update_action_dict.get("value")
        self.old_values = update_action_dict.get("old_values")
        self.transaction = update_action_dict.get("transaction", self.transaction)

        # Always remove 'history' key (if found). This prevents nested "history"
        # attributes when a project dict is loaded.
        try:
            if isinstance(self.values, dict) and "history" in self.values:
                self.values.pop("history", None)
            if isinstance(self.old_values, dict) and "history" in self.old_values:
                self.old_values.pop("history", None)
        except Exception as ex:
            log.warning('Failed to clear history attribute from undo/redo data. {}'.format(ex))


class UpdateManager:
    """ This class is used to track and distribute changes to listeners.
    Typically, only 1 instance of this class is needed, and many different
    listeners are connected with the add_listener() method. """

    def __init__(self):
        self.statusWatchers = []  # List of watchers
        self.updateListeners = []  # List of listeners
        self.actionHistory = []  # List of actions performed to current state
        self.redoHistory = []  # List of actions undone
        self.currentStatus = [None, None]  # Status of Undo and Redo buttons (true/false for should be enabled)
        self.ignore_history = False  # Ignore saving actions to history, to prevent a huge undo/redo list
        self.last_action = None  # The last action processed
        self.pending_action = None  # Last action not added to actionHistory list
        self.transaction_id = None  # The current transaction id to be attached to any UpdateActions created
        self.data_version = 0  # Incremented on every dispatch to invalidate caches
        self._is_processing_history = False  # Prevent re-entrant undo/redo from mutating history mid-iteration

    def load_history(self, project):
        """Load history from project"""
        self.reset()

        # Get history from project data
        history = project.get("history")

        # Loop through each, and load serialized data into updateAction objects
        # Ignore any load actions or history update actions
        for actionDict in history.get("redo", []):
            action = UpdateAction()
            action.load_json(json.dumps(actionDict))
            if action.type != "load" and action.key[0] != "history":
                self.redoHistory.append(action)
            else:
                log.info("Loading redo history, skipped key: %s" % str(action.key))
        for actionDict in history.get("undo", []):
            action = UpdateAction()
            action.load_json(json.dumps(actionDict))
            if action.type != "load" and action.key[0] != "history":
                self.actionHistory.append(action)
            else:
                log.info("Loading undo history, skipped key: %s" % str(action.key))

        # Notify watchers of new status
        self.update_watchers()

    def save_history(self, project, history_length):
        """Save history to project"""
        redo_list = []
        undo_list = []

        # Loop through each updateAction object and serialize
        # Ignore any load actions or history update actions
        history_length_int = int(history_length)
        if history_length_int == 0:
            self.update_untracked(["history"], {"redo": [], "undo": []})
            return
        for action in self.redoHistory[-history_length_int:]:
            if action.type != "load" and action.key[0] != "history":
                actionDict = json.loads(action.json(), strict=False)
                redo_list.append(actionDict)
            else:
                log.info("Saving redo history, skipped key: %s" % str(action.key))
        for action in self.actionHistory[-history_length_int:]:
            if action.type != "load" and action.key[0] != "history":
                actionDict = json.loads(action.json(), strict=False)
                undo_list.append(actionDict)
            else:
                log.info("Saving undo history, skipped key: %s" % str(action.key))

        # Set history data in project
        self.update_untracked(["history"], {"redo": redo_list, "undo": undo_list})

    def reset(self):
        """ Reset the UpdateManager, and clear all UpdateActions and History.
        This does not clear listeners and watchers. """
        self.actionHistory.clear()
        self.redoHistory.clear()
        self.pending_action = None
        self.last_action = None
        self._is_processing_history = False

        # Notify watchers of new history state
        self.update_watchers()

    def disconnect_listener(self, listener):
        """Remove a listener from the update manager"""
        if listener in self.updateListeners:
            log.info("Remove listener from UpdateManager: %s" % str(listener))
            self.updateListeners.remove(listener)

    def add_listener(self, listener, index=-1):
        """ Add a new listener (which will invoke the changed(action) method
        each time an UpdateAction is available). """

        if listener not in self.updateListeners:
            if index <= -1:
                # Add listener to end of list
                self.updateListeners.append(listener)
            else:
                # Insert listener at index
                self.updateListeners.insert(index, listener)
        else:
            log.warning("Cannot add existing listener: {}".format(str(listener)))

    def add_watcher(self, watcher):
        """ Add a new watcher (which will invoke the updateStatusChanged() method
        each time a 'redo' or 'undo' action is available). """

        if watcher not in self.statusWatchers:
            self.statusWatchers.append(watcher)
        else:
            log.warning("Cannot add existing watcher: {}".format(str(watcher)))

    def update_watchers(self):
        """ Notify all watchers if any 'undo' or 'redo' actions are available. """

        new_status = (len(self.actionHistory) >= 1, len(self.redoHistory) >= 1)
        if self.currentStatus[0] != new_status[0] or self.currentStatus[1] != new_status[1]:
            for watcher in self.statusWatchers:
                watcher.updateStatusChanged(*new_status)

    # This can only be called on actions already run,
    # as the old_values member is only populated during the
    # add/update/remove task on the project data store.
    # the old_values member is needed to reverse the changes
    # caused by actions.
    def get_reverse_action(self, action):
        """ Convert an UpdateAction into the opposite type (i.e. 'insert' becomes an 'delete') """
        reverse = UpdateAction(action.type, action.key, action.values)
        # On adds, setup remove
        if action.type == "insert":
            reverse.type = "delete"

            # replace last part of key with ID (so the delete knows which item to delete)
            id = action.values["id"]
            action.key.append({"id": id})

        # On removes, setup add with old value
        elif action.type == "delete":
            reverse.type = "insert"
            # Remove last item from key (usually the id of the inserted item)
            if reverse.type == "insert" and isinstance(reverse.key[-1], dict) and "id" in reverse.key[-1]:
                reverse.key = reverse.key[:-1]

        # On updates, just swap the old and new values data
        # Swap old and new values
        reverse.old_values = json.loads(json.dumps(action.values))
        reverse.values = json.loads(json.dumps(action.old_values))

        return reverse

    def undo(self):
        """ Undo the last UpdateAction (and notify all listeners and watchers).
            Continue until all identical transaction ids have been found. """
        if self._is_processing_history:
            log.warning("Ignoring re-entrant undo() while history processing is active.")
            return
        if not self.actionHistory:
            return

        self._is_processing_history = True
        try:
            # Get all actions with the same transaction id as the newest one.
            # This intentionally supports non-contiguous history entries caused by
            # delayed async operations that complete later.
            last_transaction = self.actionHistory[-1].transaction
            transaction_indexes = [
                i for i, action in enumerate(self.actionHistory)
                if action.transaction == last_transaction
            ]
            last_transactions = [self.actionHistory[i] for i in reversed(transaction_indexes)]
            remove_selection = any(a.type == "insert" for a in last_transactions)

            if remove_selection:
                # Remove selections for any items about to be deleted
                for action in last_transactions:
                    if action.type == "insert":
                        object_id = action.values.get("id", None)
                        get_app().window.removeSelection(object_id, None)

                # Force property and selection timers to fire
                get_app().window.show_property_timer.stop()
                get_app().window.show_property_timer.timeout.emit()
                get_app().window.selection_timer.stop()
                get_app().window.selection_timer.timeout.emit()
                get_app().processEvents()

            # Iterate each action in this transaction
            for index, history_index in enumerate(reversed(transaction_indexes)):
                # Remove by index to support non-contiguous matching transaction IDs.
                last_action = self.actionHistory.pop(history_index)

                # Copy action
                last_action = last_action.copy()

                # Add action to redo list
                self.redoHistory.append(last_action)
                self.pending_action = None
                # Get reverse of last action and perform it
                reverse = self.get_reverse_action(last_action)

                # Ignore updates to UI on all actions except last one
                ignore_refresh = (index != len(last_transactions) - 1)
                get_app().window.IgnoreUpdates.emit(ignore_refresh, True)

                # Perform next undo action
                self.dispatch_action(reverse)

                # Verify selections are still valid objects
                get_app().window.verifySelections()
        finally:
            self._is_processing_history = False

    def redo(self):
        """ Redo the last UpdateAction (and notify all listeners and watchers).
            Continue until all identical transaction ids have been found. """
        if self._is_processing_history:
            log.warning("Ignoring re-entrant redo() while history processing is active.")
            return
        if not self.redoHistory:
            return

        self._is_processing_history = True
        try:
            # Get all actions with the same transaction id as the newest one.
            # This intentionally supports non-contiguous history entries caused by
            # delayed async operations that complete later.
            last_transaction = self.redoHistory[-1].transaction
            transaction_indexes = [
                i for i, action in enumerate(self.redoHistory)
                if action.transaction == last_transaction
            ]
            last_transactions = [self.redoHistory[i] for i in reversed(transaction_indexes)]

            # Iterate through each action in this transaction
            for index, history_index in enumerate(reversed(transaction_indexes)):
                # Remove by index to support non-contiguous matching transaction IDs.
                next_action = self.redoHistory.pop(history_index)

                # Copy action
                next_action = next_action.copy()

                # Remove ID from insert (if found)
                if next_action.type == "insert" and isinstance(next_action.key[-1], dict) and "id" in next_action.key[-1]:
                    next_action.key = next_action.key[:-1]

                self.actionHistory.append(next_action)
                self.pending_action = None

                # Ignore updates to UI on all actions except last one
                ignore_refresh = (index != len(last_transactions) - 1)
                get_app().window.IgnoreUpdates.emit(ignore_refresh, True)

                # Perform next redo action
                self.dispatch_action(next_action)
        finally:
            self._is_processing_history = False

    # Carry out an action on all listeners
    def dispatch_action(self, action):
        """ Distribute changes to all listeners (by calling their changed() method) """

        # Every dispatched action reflects a project mutation; bump the version
        self.data_version += 1

        log.debug(
            "Dispatch action: type=%s key=%s ignore_history=%s transaction=%s",
            action.type,
            action.key,
            self.ignore_history,
            action.transaction,
        )

        try:
            # Loop through all listeners
            for listener in self.updateListeners:
                # Invoke change method on listener
                listener.changed(action)

        except Exception as ex:
            log.error("Couldn't apply '{}' to update listener: {}\n{}".format(action.type, listener, ex))
        self.update_watchers()

    # Perform load action (loading all project data), clearing history for taking a new path
    def load(self, values, reset_history=True):
        """ Load all project data via an UpdateAction into the UpdateManager
        (this action will then be distributed to all listeners) """

        self.last_action = UpdateAction('load', '', values)
        if reset_history:
            self.redoHistory.clear()
            self.actionHistory.clear()
        self.pending_action = None
        self.dispatch_action(self.last_action)

    # Perform new actions, clearing redo history for taking a new path
    def insert(self, key, values):
        """ Insert a new UpdateAction into the UpdateManager
        (this action will then be distributed to all listeners) """

        self.last_action = UpdateAction('insert', key, values, transaction=self.transaction_id)
        if self.ignore_history:
            self.pending_action = self.last_action
        else:
            self.redoHistory.clear()
            self.pending_action = None
            self.actionHistory.append(self.last_action)
        self.dispatch_action(self.last_action)

    def update(self, key, values):
        """ Update the UpdateManager with an UpdateAction
        (this action will then be distributed to all listeners) """

        self.last_action = UpdateAction('update', key, values, transaction=self.transaction_id)
        if self.ignore_history:
            self.pending_action = self.last_action
        else:
            if self.last_action.key and self.last_action.key[0] != "history":
                # Clear redo history for any update except a "history" update
                self.redoHistory.clear()
            self.pending_action = None
            self.actionHistory.append(self.last_action)
        self.dispatch_action(self.last_action)

    def update_untracked(self, key, values):
        """ Update the UpdateManager with an UpdateAction, without creating
        a new entry in the history table
        (this action will then be distributed to all listeners) """
        previous_ignore = self.ignore_history
        previous_pending = self.pending_action
        self.ignore_history = True
        self.update(key, values)
        self.ignore_history = previous_ignore
        self.pending_action = previous_pending

    def delete(self, key):
        """ Delete an item from the UpdateManager with an UpdateAction
        (this action will then be distributed to all listeners) """

        self.last_action = UpdateAction('delete', key, transaction=self.transaction_id)
        if self.ignore_history:
            self.pending_action = self.last_action
        else:
            self.redoHistory.clear()
            self.pending_action = None
            self.actionHistory.append(self.last_action)
        self.dispatch_action(self.last_action)

    def apply_last_action_to_history(self, previous_value):
        """ Apply the last action to the history """
        if self.pending_action:
            self.pending_action.set_old_values(previous_value)
            self.actionHistory.append(self.pending_action)
            self.last_action = self.pending_action
            self.pending_action = None

            # Notify watchers of new history state
            self.update_watchers()
